import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import scipy
from girard import sampling


def all_coordinates_are_positive(vec):
    return all(map(lambda pos: pos >= 0, vec))


def estimate_solid_angle(spanning_matrix, sample_size):
    dim = len(spanning_matrix)
    inverse = np.linalg.inv(spanning_matrix)

    points_inside_cone = 0

    for i in range(sample_size):
        sample_point = np.matrix(sampling.sample_hypersphere_point(dim)).T
        transformed_point = inverse * sample_point
        if all_coordinates_are_positive(transformed_point):
            points_inside_cone += 1

    return points_inside_cone / sample_size


def get_sample_std(omega, N):
    return np.sqrt(omega * (1 - omega) / N)


def get_confidence_interval_prediction(omega, N):
    std = get_sample_std(omega, N)
    return omega - 2*std, omega + 2*std


def get_sample_distribution_for_solid_angle(cone_vectors, samples_per_estimate, population_size):
    estimators = [estimate_solid_angle(cone_vectors, samples_per_estimate) for i in range(population_size)]

    mean = np.mean(estimators)
    real_std = np.std(estimators)
    predicted_std = get_sample_std(mean, samples_per_estimate)

    return estimators, mean, real_std, predicted_std



from matplotlib import pyplot as plt

def plot_multiple_functions(xrange, fn, parameters, title, xlabel, ylabel, legend_title):
    plt.yscale("log")
    plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    for p in parameters:
        plt.plot(xrange, fn(p, xrange), label=str(p))
    plt.legend(loc="upper right", title=legend_title)
    plt.show()


def plt_dim(d, b=50):
    eight = df_its[df_its["dimension"] == d]
    eight = eight["its_to_convergs"]
    m = max(eight)
    xt = 10*(m // b)
    plt.hist(eight, bins=b, density=False)
    plt.xticks(range(1, m, xt))
    plt.show()


def plot_distribution(estimators, mean, n_bins=25):
    ax = sns.distplot(estimators, fit=scipy.stats.norm, bins=n_bins, kde=False, hist=True, color='g', norm_hist=True)
    ax.set_title("Distribution for solid angle estimator")
    plt.legend(["Sample distributed", "Predicted Distribution"])
    plt.axvline(mean, 120)
    plt.show()


def build_and_plot_distribution(cone_vectors, samples_per_estimate, population_size, n_bins=25):
    estimators, mean, rstd, pstd = get_sample_distribution_for_solid_angle(cone_vectors, samples_per_estimate, population_size)
    plot_distribution(estimators, n_bins)
    return  estimators, mean, rstd, pstd


def create_plot(estimators, real_value, breaks=13, n_bins=21):
    mean = np.mean(estimators)
    std = np.std(estimators)
    ax = sns.distplot(estimators, fit=scipy.stats.norm, bins=n_bins, kde=False, hist=True, color='#e3ace3', norm_hist=True)
    ax.axvline(mean, color='#c46745', drawstyle='steps-mid', linestyle='--')
    ax.axvline(real_value, color='#69bcc2', drawstyle='steps-mid', linestyle='--')
    plt.title("Distribution of estimators for $\omega_K$")
    ax.legend(['Fitted Distribution', 'Mean', 'Acutal Solid Angle', 'Sampled Distribution'])
    plt.xlabel("Estimate of $\omega_K$")
    plt.ylabel("Estimate frequency")
    plt.xticks(np.linspace(mean - 3*std, mean + 3*std, breaks))
    plt.show()


def plot_experiments(df, dimensions):
    plt.xticks(dimensions)
    plt.xlabel("Dimension")
    plt.yscale("log")
    plt.ylabel("Samples Until Finding A Cone in Region of Convergence")
    sns.lineplot(data=df, markers=True,)
    plt.show()



import numpy as np

from decimal import Decimal
from math import sqrt, pi
from numpy import dot
from toolz import reduce

from datetime import datetime

from girard import linear_algebra as la, memoize as mem

def sampleHyperspherePoint(dim):
    gaussianSamples = np.random.normal(0, 1, dim)
    return la.normalize(gaussianSamples)


def gamma_n_plus_1_over_2(n):
    if n % 2 == 1:
        return mem.factorial((n - 1) // 2)
    else:
        sqrt_pi = Decimal(sqrt(pi))
        if n == 0:
            return sqrt_pi

        half_n = n // 2
        return (sqrt_pi / mem.power_of_two(half_n)) * mem.odd_factorial(n)


def sum_of_entries_indexed_by_i(m_tuple, n, i):
    return sum([la.access_vector_dim_n_choose_two(m_tuple, n, i, j) for j in range(n) if i != j])


#given a vector m in R^\binom{n}{k}, given as (m_11,m_12,...,m_1n, m_23,...,m_n-1,n), compute the following product of gamma functions:
# We say the entries indexed by i are the ones with format m_ij, for j != i.
# Let s_i be the sum of the entries indexed by i.
# The function computes \prod_{i = 1}^n \Gamma((s_i + 1)/2)
def gamma_product(m_tuple, n):
    sum_of_indexed_entries_by_position = [sum_of_entries_indexed_by_i(m_tuple, n, i) for i in range(n)]
    product_of_gammas = reduce(lambda x, y: x * y, map(gamma_n_plus_1_over_2, sum_of_indexed_entries_by_position))
    return Decimal(product_of_gammas)


# Explores the space of natural tuples in dimension d as a graph.
# Breadth-First Search is used to traverse the graph
def generate_tuples(d, k):
    zero = tuple(0 for i in range(d))
    tuples = set()

    tuples.add(zero)
    to_process = [zero]

    number_of_tuples = mem.factorial(d + k)/(mem.factorial(k) * mem.factorial(d))
    current_tuples = 1

    while current_tuples < number_of_tuples:
        t = to_process.pop(0)

        for pos, val in enumerate(t):
            derived_tuple = t[:pos] + (val + 1,) + t[pos + 1:]

            if derived_tuple in tuples:
                continue

            tuples.add(derived_tuple)
            to_process.append(derived_tuple)
            current_tuples += 1

    return tuples



import numpy as np


def convergence_matrix(spanning_matrix):
    grammian_matrix = spanning_matrix.T * spanning_matrix
    cm = (-1) * grammian_matrix
    np.fill_diagonal(cm, 1)
    return cm


def check_convergence(spanning_matrix):
    grammian_matrix = spanning_matrix.T * spanning_matrix
    convergence_matrix = (-1) * grammian_matrix
    np.fill_diagonal(convergence_matrix, 1)
    convergence_matrix_eigenvalues = np.linalg.eigvals(convergence_matrix)
    return min(convergence_matrix_eigenvalues) > 0



#!/usr/bin/env python
import os
from typing import List

from setuptools import setup, find_packages

# Package name used to install via pip (shown in `pip freeze` or `conda list`)
MODULE_NAME = 'girard'

# How this module is imported in Python (name of the folder inside `src`)
MODULE_NAME_IMPORT = 'girard'

# Repository name
REPO_NAME = 'girard'

# File to get direct dependencies from (used by pip)
REQUIREMENTS_FILE = 'requirements.txt'


def get_version() -> str:
    with open(os.path.join('src', MODULE_NAME_IMPORT, 'resources',
                           'VERSION')) as f:
        return f.read().strip()


def requirements_from_pip(filename: str) -> List[str]:
    with open(filename, 'r') as pip:
        return [l.strip() for l in pip if not l.startswith('#') and l.strip()]

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

SETUP_ARGS = {
    'name': MODULE_NAME,
    'description': 'Python package for approximating solid angles in high dimensions',
    'long_description': long_description,
    'long_description_content_type': "text/markdown",
    'url': 'https://github.com/gnsantos/{:s}'.format(REPO_NAME),
    'author': 'Gervásio Protásio dos Santos Neto',
    'package_dir': {'': 'src'},
    'packages': find_packages('src'),
    'version': get_version(),
    'install_requires': requirements_from_pip(REQUIREMENTS_FILE),
    'extras_require': {
        'test_deps': requirements_from_pip('requirements_test.txt')},
    'include_package_data': True,
    'zip_safe': False,
    'classifiers': [
        'Programming Language :: Python :: 3.6']
}

if __name__ == '__main__':
    setup(**SETUP_ARGS)



import numpy as np

from math import pi, sqrt, exp
from functools import reduce
from numpy import matrix, arccos
from numpy.linalg import inv, det
from decimal import *

from girard import linear_algebra as la, memoize as mem, utils

#term of the hypergeometric series in dimension \binom{n}{k}.
# a and alpha are vectors in dimension \binom{n}{k}
def series_term(m_tuple, alpha, n):
    alpha_to_the_m = la.vector_vetor_exp(alpha, m_tuple)
    sum_entries = sum(m_tuple)
    sign = (-1) ** sum_entries
    power = mem.power_of_two(sum_entries)
    fact = reduce(lambda a,b: a*b, map(mem.factorial, m_tuple))
    gamma_prod = utils.gamma_product(m_tuple, n)
    return (alpha_to_the_m * sign * power * gamma_prod) / fact

# function that puts together all previous components to implement Ribando"s formula for the solid angle
def solid_angle(spanning_matrix, max_weight):
    cone_vectors = spanning_matrix.T.tolist()
    cone_vectors = list(map(la.normalize, cone_vectors))
    n = len(cone_vectors)
    alpha = la.pairwise_dot_products(cone_vectors)
    v_matrix = matrix(cone_vectors).T

    det_V = abs(det(v_matrix))

    pi_term = (4*pi)**(n/2.0)

    constant_term = Decimal(det_V / pi_term)
    series_dimension = int(n*(n-1)/2)
    inputs = utils.generate_tuples(series_dimension, max_weight)
    series = sum(map(lambda m: series_term(m, alpha, n), inputs))
    return constant_term * series



from decimal import Decimal
from numpy import dot
from numpy.linalg import norm
from toolz import reduce

#calculates the vector,vector exponentiations:
#for a, m in R^d, a^m is:
# \prod_{i = 0}^d a_i ^ m_i
def vector_vetor_exp(alpha, m):
    return Decimal(reduce(lambda a1, a2: a1 * a2, map(lambda a_m: a_m[0]**a_m[1], zip(alpha, m))))


#normalizes a vector v
def normalize(v):
    two_norm = norm(v)
    return list(map(lambda x: x / two_norm, v))


# given a vector m in R^\binom{n}{2}, given as (m_01; m_02; ...; m_0,n-1; m_12; ...; m_n-2,n-1), acesses the positions m_ij.
def access_vector_dim_n_choose_two(m, n, i, j):
    assert i < n, "Positions must be smaller than the dimension"
    assert j < n, "Positions must be smaller than the dimension"
    assert i != j, "Not defined when i == j"

    if i > j:
        return access_vector_dim_n_choose_two(m, n, j, i)
    l = int((n - 1) * i - i*(i-1)/2)
    return m[l + j - i -1]


# compute the pairwise dot products (the dihedral angles) between the spanning vectors of the cone
def pairwise_dot_products(cone_vectors):
    n = len(cone_vectors)
    return [dot(cone_vectors[i], cone_vectors[j]) for i in range(n) for j in range(i + 1, n) if i != j]
